#include <stdio.h>
#include <stdlib.h>

#define ART_ADR     0x012478
#define MAP_16_ADR  0x00B478
#define MAP_128_ADR 0x002478
#define LEVEL_ADR   0x00A478
#define PALETTE_ADR 0x000112

#define MAX_TILE 1024
#define TYPE unsigned short
#define GEN_STATE_SIZE 138*1024

#define GRAY_LEVEL1 5*5
#define GRAY_LEVEL2 5*8
#define GRAY_LEVEL3 5*15

typedef struct
{
 unsigned short T[16][16];
 char flag;     // Reserve ?
 char palette;  // Quelle palette ?
 char fliph;    // On demande en fliph ?
 char flipv;    // On demande en flipv ?
 char fliphv;   // On demande en fliphv ?
} tile16;

char GenPalette[4][16];

tile16 sprite[MAX_TILE];

void init()
{
 int i,j,k;
 for(i = 0 ; i < MAX_TILE ; i++)
       {
        sprite[i].flag = 0;
        sprite[i].palette = 0;
        sprite[i].fliph = 0;
        sprite[i].flipv = 0;
        sprite[i].fliphv = 0;
        for(j=0;j<16;j++) for(k=0;k<16;k++) (sprite[i].T)[j][k] = 0;
       }
}

/*
 * Lecture des tiles utilises et generation de la carte
 */
TYPE* read_tile(unsigned char *gsv, int sizex, int sizey)
{
 TYPE *mat = malloc(sizeof(TYPE) * 64 * sizex * sizey);
 unsigned char *level,*m128,*t16;
 int i,j,k,m,n;
 int arg;

 level = gsv + LEVEL_ADR;

 for(j=0 ; j<(sizey * 8) ; j+=8)
    for(i = 0 ; i < (sizex*8) ; i+=8)
        {
         k = *(level++);
         t16 = gsv + MAP_128_ADR + k*128 + 1;
         for ( n = 0 ; n < 8 ; n++)
             for (m = 0 ; m < 8 ; m++)
                 {
                  arg = *(t16-1) & 3;
                  arg = arg << 8;
                  arg += *(t16);  // Vrai valeur mais depasse un octet !
                  sprite[arg].flag = 1;
                  sprite[arg].palette = (*(t16-1) >> 4) & 3;
                  switch ((*(t16-1) >>2) & 3)
                         {
                          case 3:
                               sprite[arg].fliphv = 1;
                               arg |= 0xC000;
                               break;
                          case 2:
                               sprite[arg].flipv = 1;
                               arg |= 0x8000;
                               break;
                          case 1:
                               sprite[arg].fliph = 1;
                               arg |= 0x4000;
                               break;
                         }
                  mat[sizex*8*(j+n) + (i+m) ] = arg; //((arg & 0x00FF) << 8) + (arg >> 8);
                  t16 +=2;
                 }
        }
 return (mat);
 }


void read_graph(unsigned char *gsv)
{
 unsigned char *t8;
 unsigned char *t16;
 int i,j,k,m,n;
 unsigned int adr;
 int flag;

 t16 = gsv + MAP_16_ADR;

 // Pour tous les tiles 16x16
 for ( i = 0 ; i < MAX_TILE ; i ++)
     {
     // Pour chaque element 8x8
     for(j = 0 ; j < 16 ; j+=8)
           for(k = 0 ; k < 16 ; k+=8)
                 {
                 // On derefenrece l'adresse puis on traite les pixels
                 adr = *(t16) & 7;
                 adr = adr << 8;
                 adr += *(t16+1);

                 t8 = gsv + ART_ADR + adr*4*8;

                 // Faudrait traiter le flipping vertical/horizontal, non ?
                 // Bit 11 et 12 je crois...
                 flag = ( ( (*t16) >> 3) & 3);
                 switch (flag)
                        {
                        case 0 :
                             for ( n = 0 ; n < 8 ; n++)
                                 for( m = 0 ; m < 8 ; m+=2)
                                      {
                                      (sprite[i].T)[j+n][k+m]   = (*(t8))      & 0x0F;
                                      (sprite[i].T)[j+n][k+m+1] = (*(t8) >> 4) & 0x0F;
                                      t8++;
                                      }
                             break;
                        case 1:
                             for ( n = 0 ; n < 8 ; n++)
                                 for( m = 6 ; m >=0 ; m-=2)
                                      {
                                      (sprite[i].T)[j+n][k+m]   = (*(t8) >> 4) & 0x0F;
                                      (sprite[i].T)[j+n][k+m+1] = (*(t8)     ) & 0x0F;
                                      t8++;
                                      }
                             break;
                        case 2:
                             for ( n = 7 ; n >=0 ; n--)
                                 for( m = 0 ; m < 8 ; m+=2)
                                      {
                                      (sprite[i].T)[j+n][k+m]   = (*(t8))      & 0x0F;
                                      (sprite[i].T)[j+n][k+m+1] = (*(t8) >> 4) & 0x0F;
                                      t8++;
                                      }
                             break;
                        case 3:
                             for ( n = 7 ; n >=0 ; n--)
                                 for( m = 6 ; m >=0 ; m-=2)
                                      {
                                      (sprite[i].T)[j+n][k+m+1] = (*(t8))      & 0x0F;
                                      (sprite[i].T)[j+n][k+m]   = (*(t8) >> 4) & 0x0F;
                                      t8++;
                                      }
                             break;
                        }
                 t16+=2;
                 }
     }

}

void max3(int *a, int *b, int *c)
{
 int t;

 if ((*a >= *b) && (*a >= *c))
    {
    if (*b >= *c) return ;
    t  = *c;
    *c = *b;
    *b = t;
    return;
    }
 max3(b,c,a); // Vraiment je me suis surpasse !
}


/*
 * Init la palette Genesis
 * Et donne pour chaque couleur la bonne sur Ti
 */
void init_palette(unsigned char *gsv)
{
 unsigned char *pal;
 int i,j,num;
 int b,g,r;

 for( num = 0 ; num<4; num++)
      {
      printf("\nPal %d:",num);
      // 4 palette de 16 couleurs
      pal = gsv + PALETTE_ADR + 16*2*num;

      // Chaque couleur est code sur 12 bits. //
      pal+=2;
      GenPalette[num][0] = 0; // La premiere est transparente.
      for(i = 1 ; i < 15 ; i++)
       {
        j = (*(pal)<<8) + *(pal+1);
        b = (j >> 8) & 15;
        g = (j >> 4) & 15;
        r = (j >> 0) & 15;
        printf("B%d G%d R%d ",b,g,r);
        // Gris "pondere"
        max3(&b,&g,&r);
        j = (3*b + 2*g + r);
        //j = 5 * b;
        if (j < GRAY_LEVEL1) GenPalette[num][i] = 3;
        else if (j < GRAY_LEVEL2) GenPalette[num][i] = 2;
        else if (j < GRAY_LEVEL3) GenPalette[num][i] = 1;
        else GenPalette[num][i] = 0;
        pal+=2;
        }
      }
 printf("\n");
}

unsigned char* convert_graph(unsigned char *gsv)
{
 unsigned char* graph = malloc(sizeof(char) * (4*16) * MAX_TILE);
 unsigned char* t16 = graph;
 int i,j,k,c;

 for(i = 0 ; i < MAX_TILE ; i++)
       {
        // Le mieux serait d'utilser la palette !
        // Pour le moment detection blanc ou noir...
        for( j = 0 ; j<16 ; j++)
             {
              c = 128;
              *(t16) = 0;
              *(t16+2) = 0;
              for( k = 0 ; k < 8 ; k++, c = c >> 1)
                   {
                    switch (GenPalette[sprite[i].palette][(sprite[i].T)[j][k]]) // ((sprite[i].T)[j][k]) & 3)
                           {
                            case 3 :
                                 *(t16)   |= c;
                                 *(t16+2) |= c;
                                 break;
                            case 2 :
                                 *(t16) |= c;
                                 break;
                            case 1 :
                                 *(t16+2) |= c;
                                 break;
                           }
                   }

              c = 128;
              *(t16+1) = 0;
              *(t16+3) = 0;
              for( k = 8 ; k < 16 ; k++, c = c >> 1)
                   {
                  switch (GenPalette[sprite[i].palette][(sprite[i].T)[j][k]]) // ((sprite[i].T)[j][k]) & 3)
                           {
                            case 3 :
                                 *(t16+1)   |= c;
                                 *(t16+3) |= c;
                                 break;
                            case 2 :
                                 *(t16+1) |= c;
                                 break;
                            case 1 :
                                 *(t16+3) |= c;
                                 break;
                           }
                   }
              t16 += 4;
              }
       }

 return graph;
}


int main(int argc, char *argv[])
{
 FILE *F;
 unsigned char *gsv,*mat;
 int sizex,sizey;
 int i,j,k,l,m,n;

 if (argc < 4)
    {
     printf("sonic2 genecyst.gs0 sizex(=256) sizey(=20) \n");
     exit(1);
    }
 sscanf(argv[2],"%d",&sizex);
 sscanf(argv[3],"%d",&sizey);

 printf("\nSonic2 level convertor to SMA...\n");
 printf("Open %s with sizex %d and sizey %d\n",argv[1],sizex,sizey);

 F = fopen(argv[1],"rb");
 gsv = malloc(sizeof(char) * GEN_STATE_SIZE);

 if ((F== NULL) || (gsv == NULL))
    {
     printf("Erreur d'allocation ou d'ouverture !\n");
     exit(1);
    }

 printf("%d bytes read from state file.\n",fread(gsv,sizeof(char),GEN_STATE_SIZE,F));
 fclose(F);


 init();
 init_palette(gsv);

 printf("Write to output.lvl\n");
 mat = (unsigned char*) read_tile(gsv,sizex,sizey);
 F = fopen("output.lvl","wb");
 fwrite(mat,sizeof(char),sizex*sizey*64,F);
 fclose(F);
 free(mat);

 printf("Create graph...\n");
 read_graph(gsv);
 mat = convert_graph(mat);
 F = fopen("graph.bin","wb");
 fwrite(mat,sizeof(char),4*16*MAX_TILE,F);
 fclose(F);
 free(mat);

 printf("---------------\n Statistique :\n---------------\n");

 for(i = 0, k = 0 , j = 0, l = 0; i < MAX_TILE ; i++)
        {
        if (sprite[i].flag != 0)
           k++;
        else
            {
            if (i < 256) j++;
            if (i < 512) l++;
            }
        }
 printf("%d tiles necessaire dont %d libres avant 256 et %d avant 512.\n Total libres : %d\n",k,j,l,(MAX_TILE-k));

 m = k;
 for(i = 0, k = 0 , j = 0, l = 0; i < MAX_TILE ; i++)
        {
        if (sprite[i].fliph != 0)
           k++;
        if (sprite[i].flipv != 0)
           j++;
        if (sprite[i].fliphv != 0)
           l++;
        }
 printf("FlipH: %d FlipV : %d FlipHV : %d\n",k,j,l);
 printf("Total tiles si non calcul de flipping : %d\n",m+k+j+l);
 printf("Taille prise pas les tiles sur Ti : %d (ou %d) octets.\n",m*64,(m+k+j+l)*64);
 exit(0);
}
